#include <fstream>
#include <iostream>
#include <stdexcept>
#include <Eigen/Dense>

using Eigen::VectorXd;
using Eigen::MatrixX4d;
using Eigen::MatrixXd;
using Eigen::Matrix;

using namespace std;

template<typename _Matrix_Type_>
bool pseudoInverse(const _Matrix_Type_ &a, _Matrix_Type_ &result, double
epsilon = std::numeric_limits<typename _Matrix_Type_::Scalar>::epsilon())
{
  if(a.rows()<a.cols())
      return false;

  Eigen::JacobiSVD< _Matrix_Type_ > svd = a.jacobiSvd(Eigen::ComputeThinU |
Eigen::ComputeThinV);

  typename _Matrix_Type_::Scalar tolerance = epsilon * std::max(a.cols(),
a.rows()) * svd.singularValues().array().abs().maxCoeff();

  result = svd.matrixV() * _Matrix_Type_( (svd.singularValues().array().abs() >
tolerance).select(svd.singularValues().
      array().inverse(), 0) ).asDiagonal() * svd.matrixU().adjoint();

    return true;
}

template<typename _Matrix_Type_>
double gr(_Matrix_Type_ b, double x1, double x2, double x3) {
    return b(0) + b(1) * x1 + b(2) * x2 + b(3) * x3;
}

// The Eigen index type
typedef Eigen::DenseIndex Index;

class matrixInput
{
public:
	matrixInput(Index numrows, Index numcols, std::istream& is_) 
		: A(numrows, numcols), is(is_)
	{
		for(Index i = 0; i < numrows; ++i)
			for(Index j = 0; j < numcols; ++j)
			{
				//double el;
				//is >> el;
				is >> A(i, j);
			}
	}

	MatrixXd getMatrix() const {return A;}
private:
	MatrixXd A;
	std::istream& is;
};

class myMatricesInput
{
public:
	myMatricesInput(std::istream& is_)
		: is(is_)
	{
		// first reads n and m, respectively, from the input
		is >> n_;
		is >> m_;

		// set the matrix sizes
		//std::cout << "Enter number of factors n:\n";
		X_.resize(n_, m_ + 1);
		//std::cout << "Enter number of dimensions m:\n";
		Y_.resize(n_);

		// read in the right m columns of X
		//std::cout << "Enter the n x m matrix:\n";
		matrixInput XInput(n_, m_, is);
		// set the right m columns of X
		X_.rightCols(m_) = XInput.getMatrix();
		// set the first column to one's
		X_.leftCols(1) = MatrixXd::Ones(n_, 1);

		// read in Y
		//std::cout << "Enter the n vector:\n";
		matrixInput YInput(n_, 1, is);
		// set Y
		Y_ = YInput.getMatrix();
	}

	Index n() const {return n_;}
	Index m() const {return m_;}
	MatrixXd X() const {return X_;}
	VectorXd Y() const {return Y_;}
private:
	Index n_, m_;
	MatrixXd X_;	// the n x (m + 1) matrix
	VectorXd Y_;	// the n column vector
	std::istream& is;
};

std::istream& getInputStream(char mode)
{

	std::ifstream* file;
	
	std::string filename;

	switch(mode)
	{
	case 'k':
          cout << "Insert dimensions n,m and then the matrices:";
		return std::cin;
	case 'f':
		
		cout << "Give the file name:";
		std::cin >> filename;
		file = new std::ifstream(filename.c_str());
		return *file;
	default:
		throw std::range_error("invalid mode");
	}

}

int main()
{
    cout << "Stuff"<< endl;

	cout << "Obtain model input from (k)eyboard or (f)ile?\n";
	char mode;
	cin >> mode;


	istream& is = getInputStream(mode);

	myMatricesInput Input(is);

	Index n, m;
	MatrixXd X;
	VectorXd Y;
	n = Input.n();
	m = Input.m();
	X = Input.X();
	Y = Input.Y();


    /* THE COMPLETE THREE VARIABLES MODEL */

    /* Variablex x1, x2, x3 */

    /* DATA t-DISTRIBUTION */

    VectorXd td(200);

    td(0) = 12.71;
    td(1) = 4.303;
    td(2) = 3.182;
    td(3) = 2.776; /* Is it a bug */
    td(4) = 2.571;
    td(5) = 2.447;
    td(6) = 2.365;
    td(7) = 2.306;
    td(8) = 2.262;
    td(9) = 2.228;
    td(10) = 2.201;
    td(11) = 2.179;
    td(12) = 2.160;
    td(13) = 2.145;
    td(14) = 2.131;
    td(15) = 2.12;
    td(16) = 2.11;
    td(17) = 2.101;
    td(18) = 2.093;
    td(19) = 2.086;
    td(20) = 2.08;
    td(21) = 2.074;
    td(22) = 2.069;
    td(23) = 2.064;
    td(24) = 2.06;
    td(25) = 2.056;
    td(26) = 2.052;
    td(27) = 2.048;
    td(28) = 2.045;
    td(29) = 2.042;

    for (int i=30;  i<60;  i++) td(i) = 2.02;
    for (int i=60;  i<120; i++) td(i) = 2.0;
    for (int i=120; i<200; i++) td(i) = 1.96;

    MatrixXd fd(3,200);
    fd(0, 0)  = 161.4; fd(0, 1)  = 18.51; fd(0, 2) = 10.13;  fd(0, 3) = 7.709;
    fd(0, 4)  = 6.608; fd(0, 5)  = 5.987; fd(0, 6) = 5.591;  fd(0, 7) = 5.318;
    fd(0, 8)  = 5.117; fd(0, 9)  = 4.965; fd(0, 10) = 4.844; fd(0, 11) = 4.747;
    fd(0, 12) = 4.667; fd(0, 13) = 4.6;   fd(0, 14) = 4.543; fd(0, 15) = 4.494;
    fd(0, 16) = 4.451; fd(0, 17) = 4.414; fd(0, 18) = 4.381; fd(0, 19) = 4.351;
    fd(0, 20) = 4.325; fd(0, 21) = 4.301; fd(0, 22) = 4.279; fd(0, 23) = 4.26;
    fd(0, 24) = 4.242; fd(0, 25) = 4.225; fd(0, 26) = 4.21;  fd(0, 27) = 4.196;
    fd(0, 28) = 4.183; fd(0, 29) = 4.171;

    for (int i=30;  i<60;  i++) fd(0, i) = 4.1;
    for (int i=60;  i<120; i++) fd(0, i) = 4.0;
    for (int i=120; i<200; i++) fd(0, i) = 3.9;

    fd(1, 0)  = 199.5; fd(1, 1)  = 19.0;  fd(1, 2)  = 9.552; fd(1, 3)  = 6.944;
    fd(1, 4)  = 5.786; fd(1, 5)  = 5.143; fd(1, 6)  = 4.737; fd(1, 7)  = 4.459;
    fd(1, 8)  = 4.256; fd(1, 9)  = 4.103; fd(1, 10) = 3.982; fd(1, 11) = 3.885;
    fd(1, 12) = 3.806; fd(1, 13) = 3.739; fd(1, 14) = 3.682; fd(1, 15) = 3.634;
    fd(1, 16) = 3.592; fd(1, 17) = 3.555; fd(1, 18) = 3.522; fd(1, 19) = 3.493;
    fd(1, 20) = 3.467; fd(1, 21) = 3.443; fd(1, 22) = 3.422; fd(1, 23) = 3.403;
    fd(1, 24) = 3.385; fd(1, 25) = 3.369; fd(1, 26) = 3.354; fd(1, 27) = 3.34;
    fd(1, 28) = 3.328; fd(1, 29) = 3.316;

    for (int i=30;  i<60;  i++) fd(1, i) = 3.2;
    for (int i=60;  i<120; i++) fd(1, i) = 3.1;
    for (int i=120; i<200; i++) fd(1, i) = 3.0;

    fd(2, 0) = 215.7;  fd(2, 1) = 19.16;  fd(2, 2) = 9.277;  fd(2, 3) = 6.591;
    fd(2, 4) = 5.409;  fd(2, 5) = 4.757;  fd(2, 6) = 4.347;  fd(2, 7) = 4.066;
    fd(2, 8) = 3.863;  fd(2, 9) = 3.708;  fd(2, 10) = 3.587; fd(2, 11) = 3.49;
    fd(2, 12) = 3.411; fd(2, 13) = 3.344; fd(2, 14) = 3.287; fd(2, 15) = 3.239;
    fd(2, 16) = 3.197; fd(2, 17) = 3.16;  fd(2, 18) = 3.127; fd(2, 19) = 3.098;
    fd(2, 20) = 3.072; fd(2, 21) = 3.049; fd(2, 22) = 3.028; fd(2, 23) = 3.009;
    fd(2, 24) = 2.991; fd(2, 25) = 2.975; fd(2, 26) = 2.96;  fd(2, 27) = 2.947;
    fd(2, 28) = 2.934; fd(2, 29) = 2.922;

    for (int i=30;  i<60;  i++) fd(2, i) = 2.8;
    for (int i=60;  i<120; i++) fd(2, i) = 2.7;
    for (int i=120; i<200; i++) fd(2, i) = 2.6;


    MatrixXd DWL(3,40);
    MatrixXd DWU(3,40);
    for (int i=0; i<15; i++) {
        DWL(2, i) = 0.71; DWU(2, i) = 1.61;
    }
    for (int i=15; i<20; i++) {
        DWL(2, i) = 0.89; DWU(2, i) = 1.55;
    }
    for (int i=20; i<25; i++) {
        DWL(2, i) = 1.02; DWU(2, i) = 1.54;
    }
    for (int i=25; i<30; i++) {
        DWL(2, i) = 1.12; DWU(2, i) = 1.54;
    }
    for (int i=30; i<40; i++) {
        DWL(2, i) = 1.25; DWU(2, i) = 1.57;
    }


    /* 1. THE COMPLETE MODEL */

	// original code start
    //const int n  = 20;
    //const int mm = 3;
    //int m  = mm;
	// original code end
	Index mm = m;

    double f   = fd(m-1, n-m-1-1); // 0-indexed
    double t2s = td(n-m-1-1);

    double dL = DWL(m-1, n-1);
    double dU = DWU(m-1, n-1);

	// original code start
    //MatrixX4d X(n, 4);
    //X <<
    //        1.,  98.2,  89.4, 3.00,
    //        1.,  98.7,  89.6, 3.00,
    //        1.,  99.2, 100.3, 3.00,
    //        1.,  99.6,  95.4, 3.25,
    //        1., 100.0, 102.7, 3.25,
    //        1., 100.0, 107.7, 3.25,
    //        1.,  99.2, 103.1, 3.25,
    //        1.,  99.2,  96.5, 3.25,
    //        1.,  99.9, 111.2, 3.25,
    //        1., 101.6,  98.8, 3.25,
    //        1., 102.2, 107.5, 3.25,
    //        1., 102.2,  97.7, 3.25,
    //        1., 100.4,  93.2, 3.25,
    //        1., 101.0,  95.7, 3.25,
    //        1., 102.2, 117.7, 3.25,
    //        1., 103.4,  96.7, 3.25,
    //        1., 104.0, 105.3, 3.50,
    //        1., 103.6, 115.7, 3.50,
    //        1., 103.1, 116.8, 3.50,
    //        1., 102.6, 109.0, 3.50;

    //VectorXd Y(n);
    //Y << 1598, 1592, 1586, 1592, 1609, 1641, 1623, 1635, 1648, 1642, 1644, 1670, 1657, 1664, 1654, 1661, 1664, 1680, 1673, 1676;

    //MatrixXd IN = Matrix<double, n, n>::Identity();
    //MatrixXd JN = Matrix<double, n, n>::Constant(n, n, 1);
	// original code end

	// new version
	MatrixXd IN = MatrixXd::Identity(n, n);
	MatrixXd JN = MatrixXd::Constant(n, n, 1.0);

    MatrixXd         Xp ;
    pseudoInverse(MatrixXd(X), Xp);
    MatrixXd XT  = X.transpose();
    MatrixXd YT  = Y.transpose();


    cout << "\nTHE COMPLETE MODEL" << endl;
    cout << "The matrix X:" << endl;
    cout << X << endl;
    cout << "The matrix Y:" << endl;
    cout << Y << endl;
//    cout << "The matrix Xp:" << endl;
//    cout << Xp << endl;

    MatrixXd V = XT*X;
    MatrixXd VAR;
    pseudoInverse(V, VAR);

    MatrixXd B = Xp*Y;
    MatrixXd BT = B.transpose();

    cout << "Degrees of Freedom :";
    const int DFT = n - 1;
    const int DFR = m;
    const int DFE = n - m - 1;

    cout << "df_R=" << DFR << "    df_E=" << DFE << "    df_T=" << DFT << endl;

    MatrixXd SSE = (YT * Y) - (BT * XT * Y);
    double sse = SSE(0,0);
    double SSEmax = sse;
    MatrixXd SST = (YT * Y) - (1/((double) n))*(YT*JN*Y);
    double sst = SST(0,0);
    MatrixXd SSR = SST-SSE;
    double ssr = SSR(0,0);

    cout << SST;

    cout << "SSR = " << ssr << "    SSE = " << sse << ",    SST = " << sst << endl;

    double msr = ssr / DFR;
    double mse = sse / DFE;
    double VARmax = mse;

    cout << "MSR =" << msr << ",   MSE = " << mse << ",   s = " << sqrt(mse) << endl;

    double ss = sqrt(mse);

    cout << "The coefficients of determination are :" << endl;
    double R2 = ssr / sst;
    double R4 = 1 - ((double)(n-1))/((double)n-m-1) * (1-R2);
    double R2max = R2;
    double R02 = 1 - (1-R2) * (1+(((double)m)/((double)n-m-1))*f);

    cout << "R2 = " << R2 << ", Adjusted R2=" << R4 << ", s2=" << ss*ss << ", Aitkin : R02=" << R02 << endl;

    double F = msr/mse;
    cout << "The statistical F value of the model is : F =" << F << endl;
    cout << "The H0 acceptance interval is : (0 , " << f << ")" << endl;

    if (F < f) {
        cout << "We accept the hypothessis H0, so our model is constant" << endl;
    } else {
        cout << "We reject the hypothessis H0, so our model is not constant" << endl;
    }

   

    VectorXd s(m+1);
    VectorXd b(m+1);
    VectorXd T(m+1);


    s = VAR.diagonal().array().sqrt() * ss;
    b = B.col(0);
    T = b.array()/s.array();

    cout << "The coefficients of the model are :" << endl;
    for (int i=0; i<=m; i++) {
        cout << "b"       << i << " = " << b(i) <<
                ",   s(b" << i << ")=" << s(i) <<
                ",   T(b" << i << ")=" << T(i) <<
                ",   with C.I. (" << b(i) - s(i) * t2s << "," << b(i) + s(i)*t2s << ")" <<
                endl;
    }

    cout << "The H0 acceptance interval for the coefficients is (Two Sided Test) :" << endl;
    for (int i=0; i<=m; i++) {
        if ((T(i) > - t2s) && (T(i) < t2s)) {
            cout << "So, we accept that b" << i << "=0" << endl;
        } else {
            cout << "So, we accept that b" << i << "!=0" << endl;
        }
    }


    cout << "\nMULTICOLLINEARITY CONTROL" << endl;

  
	MatrixXd xx = X.rightCols(m);
 
    VectorXd mo(m);
    VectorXd ta(m);

    for (int j=0; j<m; j++) {
        mo(j) = xx.col(j).mean();
        ta(j) = sqrt((xx.col(j).array() - mo(j)).square().sum() / (double)(n-1));
    }

    MatrixXd Xstar(n, m);
    for (int i=0; i<n; i++)
        for (int j=0; j<m; j++) {
            Xstar(i,j) = (xx(i, j) - mo(j))/(sqrt((double)(n-1)) * ta(j));
        }

    MatrixXd XTstar;
    XTstar = Xstar.transpose();

    MatrixXd R;
    R = XTstar*Xstar;

    MatrixXd S;
    pseudoInverse(R, S);

    cout << "\nThe Mean's and Std's of variables: " << endl ;
    for (int j=0; j<m; j++) {
        cout << "_x_ " << j+1 << "=" << mo(j) <<
                ", sd" << j+1 << "=" << ta(j) << endl;
    }

    cout << "\nThe Matrix X* is :" << endl;
    cout << Xstar;

    cout << "\nThe Correlation Symmetric Matrix R=(r_{ij}) with r_{ij}=r(bi, bj)" << endl;
    cout << R;

    cout << "\nThe Eigenvalues of R are : " << endl;
    cout << R.eigenvalues();

    cout << "\nThe matrix S is the Inver of R and VIF are the diagonals of S : " << endl;
    cout << S;

    cout << "\nANALYSIS OF RESIDUALS" << endl;

    VectorXd yhat(n);
    for (int i=0; i<n; i++)
        yhat(i) = gr(b, X(i, 1), X(i, 2), X(i, 3));

    VectorXd y(n);
    y = Y.col(0);

    VectorXd res(n);
    res = y - yhat;

    VectorXd sres(n);
    sres = res / ss;

    cout << "\nThe Durbin Watson statistic" << endl;
    double d = 0;
    for (int i=1; i<n; i++) {
        d+=(res(i) - res(i-1)) *
                (res(i) - res(i-1));
    }
    d/= res.array().square().sum();

    cout << "dL = "     << dL   << ", dU = "   << dU <<
            ", 4-dU = " << 4-dU << ", 4-dL = " << 4-dL <<
            ", and D-B d = " << d << endl;

    if (d < dL)
        cout << "There is positive autocorrelation." << endl;
    if ((d > dL) && (d < dU))
        cout << "No conclusion." << endl;
    if ((d > dU) && (d < 4-dU))
        cout << "We accept Ho." << endl;
    if ((d > 4 - dU) && (d < 4-dL))
        cout << "No conclusion." << endl;
    if (d > 4-dL)
        cout << "There is negative autocorrelation." << endl;


    MatrixXd H;
    H = X*VAR*XT;

    VectorXd h;
    h = H.diagonal();

    VectorXd r(n);
    for (int i=0; i<n; i++)
         r(i) = res(i) / (ss * sqrt(1-h(i)));

    VectorXd sss(n);
    for (int i=0; i<n; i++)
        sss(i) = sqrt(((n-m)*ss*ss-(res(i)*res(i)/(1-h(i))))/(n-m-1));
    VectorXd t(n);
    for (int i=0; i<n; i++)
        t(i) = res(i) /(sss(i) * sqrt(1-h(i)));

    cout << "\nThe Matrix of Residuals etc";
    for (int i=0; i<n; i++) {
        cout << "y" << i << "=" << y(i) <<
                ", yhat_" << i << "=" << yhat(i) <<
                ", e" << i << "=" << res(i) <<
                ", e" << i << "/s=" << sres(i) <<
                ", h" << i << "=" << h(i) <<
                ", r" << i << "=" << r(i) <<
                ", t" << i << "=" << t(i) << endl;
    }

    cout << "The H0 acceptance interval for the R-students is (Two Sided Test) :" << endl;
    cout << "(" << -t2s << " , " << t2s << ")" << endl;

    for (int i=0; i<n; i++) {
        if ((t(i) > - t2s) && (t(i) < t2s)) {
            cout << "So, we have that t" << i << " = " << t(i) << " is OK" << endl;
        } else {
            cout << "So, we have that t" << i << " = " << t(i) << " is NOT OK" << endl;
        }
    }

    /* 2. WITH ONE VARIABLE xi MODELS */

    if ( mm > 1) {
        m   = 1;
        f   = fd(m-1, n-m-1-1);
        t2s = td(n-m-1-1);
    }

    return 0;
}

